//
// Created by Attila Haluska on 8/2/18.
//

#include "debug/debug.h"

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "compiler/lexer/token.h"
#include "compiler/parser/ast/ast_node.h"
#include "memory_management/memory.h"

static char *debug_token_type(const Token *token) {
    switch (token->type) {
        case TOKEN_LEFT_PAREN:
            return "TOKEN_LEFT_PAREN";
        case TOKEN_RIGHT_PAREN:
            return "TOKEN_RIGHT_PAREN";
        case TOKEN_LEFT_BRACE:
            return "TOKEN_LEFT_BRACE";
        case TOKEN_RIGHT_BRACE:
            return "TOKEN_RIGHT_BRACE";
        case TOKEN_COMMA:
            return "TOKEN_COMMA";
        case TOKEN_DOT:
            return "TOKEN_DOT";
        case TOKEN_MINUS:
            return "TOKEN_MINUS";
        case TOKEN_MINUS_MINUS:
            return "TOKEN_MINUS_MINUS";
        case TOKEN_PLUS:
            return "TOKEN_PLUS";
        case TOKEN_PLUS_PLUS:
            return "TOKEN_PLUS_PLUS";
        case TOKEN_SEMICOLON:
            return "TOKEN_SEMICOLON";
        case TOKEN_SLASH:
            return "TOKEN_SLASH";
        case TOKEN_STAR:
            return "TOKEN_STAR";
        case TOKEN_QUESTION:
            return "TOKEN_QUESTION";
        case TOKEN_COLON:
            return "TOKEN_COLON";
        case TOKEN_BANG:
            return "TOKEN_BANG";
        case TOKEN_BANG_EQUAL:
            return "TOKEN_BANG_EQUAL";
        case TOKEN_EQUAL:
            return "TOKEN_EQUAL";
        case TOKEN_EQUAL_EQUAL:
            return "TOKEN_EQUAL_EQUAL";
        case TOKEN_GREATER:
            return "TOKEN_GREATER";
        case TOKEN_GREATER_EQUAL:
            return "TOKEN_GREATER_EQUAL";
        case TOKEN_LESS:
            return "TOKEN_LESS";
        case TOKEN_LESS_EQUAL:
            return "TOKEN_LESS_EQUAL";
        case TOKEN_IDENTIFIER:
            return "TOKEN_IDENTIFIER";
        case TOKEN_STRING:
            return "TOKEN_STRING";
        case TOKEN_NUMBER:
            return "TOKEN_NUMBER";
        case TOKEN_AND_AND:
            return "TOKEN_AND_AND";
        case TOKEN_CLASS:
            return "TOKEN_CLASS";
        case TOKEN_ELSE:
            return "TOKEN_ELSE";
        case TOKEN_FALSE:
            return "TOKEN_FALSE";
        case TOKEN_FUNC:
            return "TOKEN_FUNC";
        case TOKEN_FOR:
            return "TOKEN_FOR";
        case TOKEN_IF:
            return "TOKEN_IF";
        case TOKEN_NIL:
            return "TOKEN_NIL";
        case TOKEN_PIPE_PIPE:
            return "TOKEN_PIPE_PIPE";
        case TOKEN_PRINT:
            return "TOKEN_PRINT";
        case TOKEN_RETURN:
            return "TOKEN_RETURN";
        case TOKEN_SUPER:
            return "TOKEN_SUPER";
        case TOKEN_THIS:
            return "TOKEN_THIS";
        case TOKEN_TRUE:
            return "TOKEN_TRUE";
        case TOKEN_VAR:
            return "TOKEN_VAR";
        case TOKEN_WHILE:
            return "TOKEN_WHILE";
        case TOKEN_ERROR:
            return "TOKEN_ERROR";
        case TOKEN_EOF:
            return "TOKEN_EOF";
    }

    return "TOKEN_UNKNOWN";
}

void debug_token(Token *token) {
    char* lexeme = NEW_WITH_SIZE(char, token->length + 1);
    memcpy(lexeme, token->start, token->length);
    lexeme[token->length] = '\0';
    fprintf(stdout, "Token{%s, '%s'}\n", debug_token_type(token), lexeme);
    DELETE_TYPE(char, lexeme);
}

static int depth = 0;
static char* indent = "  ";

static void begin_scope() {
    depth++;
}

static void end_scope() {
    depth--;
}

static void print_with_indent(const char* what, int extra_args, ...) {
    for (int i = 0; i < depth; i++) {
        fprintf(stdout, "%s", indent);
    }

    fprintf(stdout, "%s", what);

    va_list args;
    va_start(args, extra_args);

    for (int i = 0; i < extra_args; i++) {
        const char* arg = va_arg(args, const char*);
        fprintf(stdout, "%s", arg);
    }

    va_end(args);

    fprintf(stdout, "\n");
}

static char* integer_to_string(int64_t number) {
    int length = snprintf(NULL, 0, "%lld", number);
    char* str = NEW_WITH_SIZE(char, length + 1);
    snprintf(str, length + 1, "%lld", number);
    return str;
}

static char* double_to_string(long double number) {
    int length = snprintf(NULL, 0, "%Lf", number);
    char* str = NEW_WITH_SIZE(char, length + 1);
    snprintf(str, length + 1, "%Lf", number);
    return str;
}

static void debug_integer_node(AstNode* node) {
    print_with_indent("IntegerExpressionNode {", 0);
    begin_scope();
    char* str = integer_to_string(node->as.integer);
    print_with_indent(str, 0);
    DELETE_TYPE(char, str);
    end_scope();
    print_with_indent("}", 0);
}

static void debug_double_node(AstNode* node) {
    print_with_indent("DoubleExpressionNode {", 0);
    begin_scope();
    char* str = double_to_string(node->as.double_precision);
    print_with_indent(str, 0);
    DELETE_TYPE(char, str);
    end_scope();
    print_with_indent("}", 0);
}
static void debug_binary_infix_node(AstNode* node) {
    print_with_indent("BinaryInfixOperatorExpressionNode {", 0);
    begin_scope();
    debug_ast_node(node->as.binaryInfix.left);
    char* str = token_get_lexeme(node->token);
    print_with_indent("Operator ", 1, str);
    DELETE_TYPE(char, str);
    debug_ast_node(node->as.binaryInfix.right);
    end_scope();
    print_with_indent("}", 0);
}

void debug_ast_node(AstNode* node) {
    switch (node->type) {
        case NODE_INTEGER:
            debug_integer_node(node);
            break;
        case NODE_DOUBLE:
            debug_double_node(node);
            break;
        case NODE_BINARY_INFIX_OPERATOR:
            debug_binary_infix_node(node);
            break;
    }
}
